@@ -5,7 +5,7 @@ from matplotlib import colors
 import matplotlib
 from matplotlib.patches import Polygon
 from shapely import geometry
-# from mpl_toolkits.basemap import Basemap
+from mpl_toolkits.basemap import Basemap
 from matplotlib.collections import PatchCollection
 # from landez import ImageExporter
 # from landez.filters import GrayScale
@@ -14,6 +14,7 @@ import itertools
 import math
 from .core import shift_down, shift_up
 from .core import movoper
+import dijkstra
 
 
 def plt_trajectory(lat, lng, trip_id=None, max_sample=100, line_style="solid", line_width=2, color="gray", marker="o", marker_size=5, mark_od=True, show_ticks=False, show=False):
@@ -410,6 +411,135 @@ def point_within(x, y, polygon):
         raise Exception("polygon_coord has to be a list of tuple (x, y)")
 
 
+class DijkstraEstimator:
+    def __init__(self, node_file_path, link_file_path, dist_col="dist_corr"):
+        self.graph3 = dijkstra.Graph()
+        self.graph1 = dijkstra.Graph()
+        links_df = pd.read_csv(link_file_path)
+        for r, row in links_df.iterrows():
+            self.graph1.add_edge(row["frm_node"], row["to_node"], row[dist_col])
+            if row["n_link_sample"] >= 3:
+                self.graph3.add_edge(row["frm_node"], row["to_node"], row[dist_col])
+        nodes_df = pd.read_csv(node_file_path)
+        self.node_lat_arr = nodes_df["lat"].values
+        self.node_lng_arr = nodes_df["lng"].values
+        self.node_names = nodes_df["node"].values
+        self.node_col = np.argwhere(nodes_df.columns.values == "node")[0][0]
+
+    def estimate_by_node(self, ori_node, dest_node):
+        dijkstra3 = dijkstra.DijkstraSPF(self.graph3, ori_node)
+        dist = dijkstra3.get_distance(dest_node)
+        if dist < math.inf:
+            path = dijkstra3.get_path(dest_node)
+            return dist, path
+        else:
+            dijkstra1 = dijkstra.DijkstraSPF(self.graph1, ori_node)
+            dist = dijkstra1.get_distance(dest_node)
+            if dist < math.inf:
+                path = dijkstra1.get_path(dest_node)
+                #                 print("Weak link:", ori_node, dest_node, "est dist", dist)
+                return dist, path
+            else:
+                return 0.0, None
+
+    def estimate_by_loc(self, ori_lat, ori_lng, dest_lat, dest_lng, alpha=1.1, cache_dic=None, google_key=""):
+        if (ori_lat is None) or (ori_lat == 0.0) or (dest_lat is None) or (dest_lat == 0.0):
+            return None, None, None, ""
+        if (ori_lat == dest_lat) and (ori_lng == dest_lng):
+            return 0, 0, "", ""
+        ori_y = self.node_lat_arr - ori_lat
+        ori_x = self.node_lng_arr - ori_lng
+        ori_nearest_dists = np.round((ori_y ** 2 + ori_x ** 2) ** 0.5 * np.pi * 6371 / 2 / 90, 3)
+        ind = np.argmin(ori_nearest_dists)
+        ori2node_dist = ori_nearest_dists[ind]
+        ori_node = self.node_names[ind]
+        dest_y = self.node_lat_arr - dest_lat
+        dest_x = self.node_lng_arr - dest_lng
+        dest_nearest_dists = np.round((dest_y ** 2 + dest_x ** 2) ** 0.5 * np.pi * 6371 / 2 / 90, 3)
+        ind = np.argmin(dest_nearest_dists)
+        dest2node_dist = dest_nearest_dists[ind]
+        dest_node = self.node_names[ind]
+        if cache_dic is not None:
+            node2node_dist_cache_key = "DIST:" + ori_node + "-" + dest_node
+            path_cache_key = "PATH:" + ori_node + "-" + dest_node
+            if cache_dic.__contains__(node2node_dist_cache_key) and cache_dic.__contains__(path_cache_key):
+                node2node_dist = cache_dic[node2node_dist_cache_key]
+                path = cache_dic[path_cache_key]
+            else:
+                node2node_dist, path = self.estimate_by_node(ori_node, dest_node)
+                cache_dic[node2node_dist_cache_key] = node2node_dist
+                cache_dic[path_cache_key] = path
+        else:
+            node2node_dist, path = self.estimate_by_node(ori_node, dest_node)
+        if path is not None:
+            total_dist = np.round(ori2node_dist * alpha + node2node_dist + dest2node_dist * alpha, 3)
+            center_lat, center_lng = np.mean([ori_lat, dest_lat]), np.mean([ori_lng, dest_lng])
+            between_markers = "".join(["|" + node[node.find("@") + 1:] for node in path])
+            if (np.abs(ori_lng - dest_lng) > 0.18) or (np.abs(ori_lat - dest_lat) > 0.18):
+                url = "https://maps.googleapis.com/maps/api/staticmap?center=Singapore&zoom=11&size=1024x768&maptype=roadmap%20&style=feature:administrative|visibility:off&style=feature:poi|visibility:off&style=feature:landscape|visibility:off&style=feature:road.arterial|visibility:off&key=" + google_key + "&markers=color:green|size:tiny|" + str(
+                    ori_lat) + "," + str(
+                    ori_lng) + "&markers=color:black|size:tiny" + between_markers + "&markers=color:red|size:tiny|" + str(
+                    dest_lat) + "," + str(dest_lng)
+            elif (np.abs(ori_lng - dest_lng) > 0.08) or (np.abs(ori_lat - dest_lat) > 0.08):
+                url = "https://maps.googleapis.com/maps/api/staticmap?center=" + str(center_lat) + "," + str(
+                    center_lng) + "&zoom=12&size=1024x768&maptype=roadmap%20&style=feature:administrative|visibility:off&style=feature:poi|visibility:off&style=feature:landscape|visibility:off&style=feature:road.arterial|visibility:off&key=" + google_key + "&markers=color:green|size:tiny|" + str(
+                    ori_lat) + "," + str(
+                    ori_lng) + "&markers=color:black|size:tiny" + between_markers + "&markers=color:red|size:tiny|" + str(
+                    dest_lat) + "," + str(dest_lng)
+            else:
+                url = "https://maps.googleapis.com/maps/api/staticmap?center=" + str(center_lat) + "," + str(
+                    center_lng) + "&zoom=13&size=1024x768&maptype=roadmap%20&style=feature:administrative|visibility:off&style=feature:poi|visibility:off&style=feature:landscape|visibility:off&style=feature:road.arterial|visibility:off&key=" + google_key + "&markers=color:green|size:small|" + str(
+                    ori_lat) + "," + str(
+                    ori_lng) + "&markers=color:black|size:small" + between_markers + "&markers=color:red|size:small|" + str(
+                    dest_lat) + "," + str(dest_lng)
+            return total_dist, node2node_dist, path, url
+        else:
+            y = ori_lat - dest_lat
+            x = ori_lng - dest_lng
+            total_dist = np.round((y ** 2 + x ** 2) ** 0.5 * np.pi * 6371 / 2 / 90 * 1.4, 3)
+            print("Fail:", ori_node, dest_node, "est dist", total_dist)
+            return total_dist, node2node_dist, path, ""
+
+    def estimate_by_loc_for_array(self, lat_a, lng_a, lat_b, lng_b, cache_dic=None):
+        output = np.zeros(lat_a.shape)
+        if len(lat_a.shape) == 1:
+            for i in range(lat_a.shape[0]):
+                dist, _, path, _ = self.estimate_by_loc(lat_a[i], lng_a[i], lat_b[i], lng_b[i], cache_dic=cache_dic)
+                output[i] = dist
+        elif len(lat_a.shape) == 2:
+            for i in range(lat_a.shape[0]):
+                for j in range(lat_a.shape[1]):
+                    dist, _, path, _ = self.estimate_by_loc(lat_a[i, j], lng_a[i, j], lat_b[i, j], lng_b[i, j],
+                                                                            cache_dic=cache_dic)
+                    output[i, j] = dist
+        elif len(lat_a.shape) == 3:
+            for i in range(lat_a.shape[0]):
+                for j in range(lat_a.shape[1]):
+                    for k in range(lat_a.shape[2]):
+                        dist, _, path, _ = self.estimate_by_loc(lat_a[i, j, k], lng_a[i, j, k], lat_b[i, j, k], lng_b[i, j, k],
+                                                                                cache_dic=cache_dic)
+                        output[i, j, k] = dist
+
+        return output
+
+    def estimate_by_nodes(self, ori_nodes, dest_nodes):
+        assert len(ori_nodes) == len(dest_nodes), "ori_nodes, dest_nodes must have same length"
+        dist_arr = np.zeros(len(ori_nodes))
+        for i in range(len(ori_nodes)):
+            ori_node = ori_nodes[i]
+            dest_node = dest_nodes[i]
+            dijkstra3 = dijkstra.DijkstraSPF(self.graph3, ori_node)
+            dist = dijkstra3.get_distance(dest_node)
+            if dist < math.inf:
+                dist_arr[i] = dist
+            else:
+                dijkstra1 = dijkstra.DijkstraSPF(self.graph1, ori_node)
+                dist = dijkstra1.get_distance(dest_node)
+                if dist < math.inf:
+                    dist_arr[i] = dist
+        return dist_arr
+
+
 
 
 # basemap = Basemap(llcrnrlon= 75,llcrnrlat=10,urcrnrlon=150,urcrnrlat=55,projection='poly',lon_0 = 116.65,lat_0 = 40.02,ax = ax)
